import java.io.File;
import java.util.Date;
import java.util.List;
import java.util.UUID;

import javax.persistence.Basic;
import javax.persistence.CascadeType;
import javax.persistence.ElementCollection;
import javax.persistence.Embeddable;
import javax.persistence.Entity;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.FlushModeType;
import javax.persistence.Id;
import javax.persistence.ManyToOne;
import javax.persistence.Persistence;
import javax.persistence.Query;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;

public final class UnexpectedMemoryUsageAfterClose {

	public static void main(String[] args) {
		System.out.println("start");

		File dbFile = new File("E:/objectdb/db/test.odb");
		if (dbFile.exists()) {
			dbFile.delete();
		}

		EntityManagerFactory emf = Persistence.createEntityManagerFactory("objectdb:$objectdb/db/test.tmp;drop");
		EntityManager em = emf.createEntityManager();
		// em.setFlushMode(FlushModeType.AUTO);

		// Persist an entity with an embeddable:
		em.getTransaction().begin();
		for (int i = 0; i < 500000; i++) {

			Message msg = new MessageImpl();
			msg.setTypeID(UUID.randomUUID().toString()
					+ "1234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890");

			if (i % 10000 == 0) {
				msg.setTypeID("" + i);
			}

			em.persist(msg);

			if (i % 10000 == 0) {
				System.out.println("count: " + i);
				em.flush();
				em.clear();
				System.out.println("flush: " + i / 10000);
			}

			if (i % 10000 == 0) {
				System.out.println("query: " + i / 10000);
				Query q = em.createQuery("Select m FROM MessageImpl m WHERE m.typeID = '" + i + "'");
				List res = q.getResultList();
				res.size();
			}
		}
		System.out.println("Add a break point... Used Heap is over 1 GB ");
		em.getTransaction().commit();
		em.close();
		em = null;

		System.out.println("Add a break point... Used Heap is over 1 GB ");
		emf.close();
		emf = null;

		System.out.println("Add a break point... Used Heap is over 1 GB ");
		// trigger GC
		System.gc();
		System.out.println("Add a break point... Over 400 MB  byte array are still referenced..");

		
		// open a new DB
		emf = Persistence.createEntityManagerFactory("objectdb:$objectdb/db/test.tmp;drop");
		em = emf.createEntityManager();
		em.setFlushMode(FlushModeType.AUTO);

		// Persist an entity with an embeddable:
		em.getTransaction().begin();

		Message msg = new MessageImpl();
		msg.setTypeID("123");
		em.persist(msg);

		Query q = em.createQuery("Select m FROM MessageImpl m WHERE m.typeID = '123'");
		List res = q.getResultList();
		res.size();

		em.getTransaction().commit();
		em.close();
		em = null;

		System.out.println("add break point");
		emf.close();
		emf = null;

		// trigger GC
		System.gc();
		System.out.println("Add a break point... Now the 400 MB not longer referenced.. !! ");
	}

	@Entity
	public static class MessageImpl extends ModelElementImpl implements Message {

		@ManyToOne(cascade = { CascadeType.REFRESH }, targetEntity = ModelElementImpl.class)
		private ModelElement causingObject;

		@Basic
		private MessageSeverity severity;

		@Basic
		private String typeID;

		@ElementCollection(targetClass = TokenImpl.class)
		private List<Token> messageTokens;

		@ElementCollection(targetClass = TokenImpl.class)
		private List<Token> hintTokens;

		@Basic
		@Temporal(TemporalType.TIMESTAMP)
		private Date date;

		/**
		 * Construct a new message and set its date to this time of
		 * instantiation.
		 */
		public MessageImpl() {
			super();
			this.date = new Date();
		}

		@Override
		public void setTypeID(String typeID) {
			this.typeID = typeID;
		}

		@Override
		public String getTypeID() {
			return this.typeID;
		}

		@Override
		public void setMessage(List<Token> messageTokens) {
			this.messageTokens = messageTokens;
		}

		/**
		 * Sets the date of creation of this message.
		 *
		 * @param created
		 *            date that this message was created.
		 */
		public void setDate(Date created) {
			this.date = created;
		}
	}

	@Entity
	public static abstract class ModelElementImpl implements ModelElement {

		@Id
		private String uid;

		/**
		 * JPA Entity Default constructor
		 * <p>
		 * The default constructor must be a public or protected zero-argument
		 * constructor.
		 * </p>
		 */
		protected ModelElementImpl() {
			this.uid = UUID.randomUUID().toString().toUpperCase().replaceAll("-", "");
		}

		@Override
		public int hashCode() {
			if (uid == null) {
				return super.hashCode();
			}

			return uid.hashCode();
		}

		@Override
		public boolean equals(Object obj) {
			if (obj instanceof ModelElement && this.getUid() != null) {
				return this.getUid().equals(((ModelElement) obj).getUid());
			}

			return super.equals(obj);
		}

		@Override
		public final String getUid() {
			return this.uid;
		}
	}

	public static interface Message extends ModelElement {

		public static interface Token {
			public String toString();
		}

		/**
		 * Get the ID of the message type of this message.
		 * 
		 * @return the ID of the message type of this message.
		 */
		public String getTypeID();

		public void setMessage(List<Token> messageTokens);

		/**
		 * Set the ID of the message type of this message.
		 * 
		 * @param typeID
		 *            the ID of the message type of this message.
		 */
		public void setTypeID(String typeID);

	}

	public static abstract interface ModelElement {

		/**
		 * Get the Universally Unique Identifier for this object.
		 * 
		 * @return Universally Unique Identifier
		 */
		public String getUid();

		/**
		 * <p>
		 * Unless specialized further by subclasses, any two ModelElement
		 * objects are equal if they have the same UID. Unless specialized
		 * further, this method is therefore guaranteed to return true whenever
		 * the given {@code obj} has the same UID as {@code this}. Subclasses
		 * may decide that ModelElement objects are equal even if they have
		 * different UIDs.
		 * </p>
		 * 
		 * <h3>Generic inherited javadoc follows.</h3> {@inheritDoc}
		 */
		@Override
		public boolean equals(Object obj);

		/**
		 * For any two ModelElements with the same non-null UID, this method is
		 * guaranteed to return the same hash code.
		 * 
		 * <h3>Generic inherited javadoc follows.</h3> {@inheritDoc}
		 */
		@Override
		public int hashCode();
	}

	public static enum MessageSeverity {
		INFO, WARNING, ERROR
	}

	@Embeddable
	public static class TokenImpl implements Token {

		@Basic
		String string;

		protected TokenImpl() {
		}
	}

	public static interface Token {

	}
}
